#!/usr/bin/env python3

"""Fragmapper analysis tools.

Description:
FragMapper compares reactivity profile differences between SHAPE-MaP profiles.
The intended application of Fragmapper is to detect fragment or ligand
crosslinking sites in RNA.
"""

import pandas as pd
import numpy as np
from scipy import stats
from statsmodels.stats.multitest import multipletests
import matplotlib.pyplot as plt
from rnavigate import data, Sample

__version__ = "0.1.0"
__author__ = "Seth D. Veenbaas"
__maintainer__ = "Seth D. Veenbaas"
__email__ = "sethv@live.unc.edu"


class FragMaP(data.Profile):
    def __init__(
        self,
        input_data,
        parameters,
        metric="Fragmap_profile",
        metric_defaults=None,
        read_table_kw=None,
        sequence=None,
        name=None,
    ):
        self.parameters = parameters
        if isinstance(input_data, (list, tuple)) and len(input_data) == 2:
            profile1, profile2 = input_data
            input_data = self.get_dataframe(profile1, profile2, **parameters)
        elif not isinstance(input_data, pd.DataFrame):
            raise ValueError(
                "Input data must be a list of 2 shapemap profiles or a "
                "pre-initialized dataframe"
            )
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Fragmap_profile": {
                "metric_column": "Fragmap_profile",
                "error_column": "Fragmap_err",
                "color_column": "Site",
                "cmap": ["lightgrey", "limegreen"],
                "normalization": "none",
                "values": None,
                "extend": "neither",
                "alpha": 1.0,
                "title": "Frag-MaP sites",
                "ticks": [0, 1],
                "tick_labels": ["other", "sites"],
            }
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            sequence=sequence,
            name=name,
        )

    @property
    def recreation_kwargs(self):
        return {"parameters": self.parameters}

    def get_dataframe(
        self,
        profile1,
        profile2,
        mutation_rate_threshold,
        depth_threshold,
        p_significant,
        ss_threshold,
        correction_method,
    ):
        columns = [
            "Nucleotide",
            "Sequence",
            "Modified_mutations",
            "Modified_effective_depth",
            "Modified_rate",
            "Std_err",
        ]
        dataframe = pd.merge(
            profile1.data[columns],
            profile2.data[columns],
            how="left",
            on=["Nucleotide", "Sequence"],
            suffixes=("_1", "_2"),
        )

        # Filter data
        dataframe["filter"] = dataframe.eval(
            # Effective read depths must be greater than threshold
            f"Modified_effective_depth_1 > {depth_threshold} "
            f"& Modified_effective_depth_2 > {depth_threshold} "
            # Control mutation rate (profile 2) must be less than threshold
            f"& Modified_rate_2 < {mutation_rate_threshold}"
        ).astype(bool)
        valid = dataframe["filter"]

        # Calculate Z-scores for profile 1 and profile 2
        dataframe[["zscore_1", "zscore_2"]] = np.nan
        # Z-scores per nucleotide
        FragMaP.calc_zscore(
            self,
            valid,
            dataframe,
            incolumn="Modified_rate_1",
            outcolumn="zscore_1",
            base=["A", "U", "C", "G"],
        )
        FragMaP.calc_zscore(
            self,
            valid,
            dataframe,
            incolumn="Modified_rate_2",
            outcolumn="zscore_2",
            base=["A", "U", "C", "G"],
        )

        # Frag-MaP profile is the difference in Z-scores
        dataframe.eval("Fragmap_profile = zscore_1 - zscore_2", inplace=True)
        dataframe.eval(
            "Fragmap_err = sqrt(zscore_1_err ** 2 + zscore_2_err ** 2)", inplace=True
        )

        dataframe["Fragmap_pvalue"] = stats.norm.sf(dataframe["Fragmap_profile"])

        # Calculate T-statistic and P-value
        dataframe["tstat"], dataframe["pvalue"] = stats.ttest_ind_from_stats(
            mean1=dataframe["Modified_rate_1"],
            std1=dataframe["Std_err_1"],
            nobs1=dataframe["Modified_effective_depth_1"],
            mean2=dataframe["Modified_rate_2"],
            std2=dataframe["Std_err_2"],
            nobs2=dataframe["Modified_effective_depth_2"],
            equal_var=False,
        )

        # Bonferoni p-value correction
        p_significant = p_significant / len(dataframe)
        dataframe["Significant"] = dataframe["pvalue"] < p_significant

        # Calculate Delta rate
        dataframe.eval("Delta_rate = Modified_rate_1 - Modified_rate_2", inplace=True)
        dataframe["Delta_rate_zscore"] = stats.zscore(dataframe["Delta_rate"])

        # Perform False Discovery Rate (FDR) correction
        if correction_method is not None:
            p_values = dataframe["Fragmap_pvalue"].values
            reject, pvals_corrected, _, _ = multipletests(
                p_values, alpha=ss_threshold, method=correction_method
            )
            dataframe["Reject"] = reject
            dataframe["Fragmap_pvalue_corrected"] = pvals_corrected
        else:
            dataframe["Reject"] = dataframe["Fragmap_pvalue"] < ss_threshold
            dataframe["Fragmap_pvalue_corrected"] = np.nan

        # Called Frag-MaP sites
        dataframe["Site"] = (dataframe["Reject"]) & (dataframe["Significant"])

        return dataframe

    def calc_zscore(
        self, valid, dataframe, incolumn: str, outcolumn: str, base: list
    ) -> None:
        sele_data = dataframe.loc[dataframe["Sequence"].isin(base)].copy()

        # Calculate outlier nucleotides
        Q1 = sele_data[incolumn].quantile(0.25)
        Q3 = sele_data[incolumn].quantile(0.75)
        IQR = Q3 - Q1
        lower_limit = Q1 - 1.5 * IQR
        upper_limit = Q3 + 1.5 * IQR
        nonoutlier = dataframe[incolumn].between(lower_limit, upper_limit)

        # Calculate zscore
        valid_nt = valid & dataframe["Sequence"].isin(base)

        sem_column = f"Std_err_{incolumn.split('_')[-1]}"
        value = np.log(dataframe.loc[valid_nt, incolumn])
        sem = np.log(dataframe.loc[valid_nt, sem_column])
        mean = np.mean(np.log(dataframe.loc[valid_nt & nonoutlier, incolumn]))
        std = np.std(np.log(dataframe.loc[valid_nt & nonoutlier, incolumn]))

        dataframe.loc[valid_nt, outcolumn] = (value - mean) / std

        # Calculate the partial derivatives
        dz_dy = (mean * np.log(value)) / std
        dz_dx = np.log(value) / std
        dz_ds = -(mean * np.log(value) / (std**2))

        # Propagate error using the error propagation formula
        # Note: The SEMs for "x" and "s" are assumed to be zero.
        fragmap_err = np.sqrt((dz_dy * sem) ** 2 + (dz_dx * 0) ** 2 + (dz_ds * 0) ** 2)

        dataframe.loc[valid_nt, f"{outcolumn}_err"] = fragmap_err

    def get_annotation(self):
        return data.Annotation(
            input_data=self.data.loc[self.data["Site"], "Nucleotide"].to_list(),
            name="Frag-MaP sites",
            sequence=self.sequence,
            annotation_type="sites",
            color="green",
        )


class Fragmapper(Sample):
    def __init__(self, sample1, sample2, parameters=None, profile="shapemap"):
        if sample1.data[profile].sequence != sample2.data[profile].sequence:
            raise ValueError("Profiles must have the same sequence")
        if parameters is None:
            parameters = {}
        self.parameters = {
            "mutation_rate_threshold": 0.025,
            "depth_threshold": 5000,
            "p_significant": 0.001,
            "ss_threshold": 0.05,
            "correction_method": None,
        }
        self.parameters |= parameters
        fragmap = FragMaP(
            input_data=[sample1.get_data(profile), sample2.get_data(profile)],
            parameters=self.parameters,
        )
        super().__init__(
            sample=f"FragMaP: {sample1.sample}/{sample2.sample}",
            inherit=[sample1, sample2],
            fragmap=fragmap,
            fragmap_sites=fragmap.get_annotation(),
        )
        self.sample1 = sample1
        self.sample2 = sample2

    def plot_scatter(self, column="Modified_rate"):
        """Generates scatter plots useful for fragmapper quality control.

        Args:
            column (str, optional):
                Dataframe column containing data to plot (must be avalible for
                the sample and control).
                Defaults to "Modified_rate".

        Returns:
            (matplotlib figure, matplotlib axis)
                Scatter plot with control values on the x-axis, sample values
                on the y-axis, and each point representing a nucleotide not
                filtered out in the fragmapper pipeline.
        """
        fig, ax = plt.subplots(1, 1, figsize=(4, 4))

        columns = [f"{column}_1", f"{column}_2"]
        scatter_data = self.data["fragmap"].data
        scatter_data = scatter_data[scatter_data["filter"]].copy()
        sites = scatter_data[scatter_data["Site"]]
        non_sites = scatter_data[~scatter_data["Site"]]

        for _, (nt, y, x) in sites[["Nucleotide"] + columns].iterrows():
            ax.text(x, y, int(nt))

        ax.scatter(x=non_sites[columns[1]], y=non_sites[columns[0]], s=5, alpha=0.25)
        ax.scatter(x=sites[columns[1]], y=sites[columns[0]], s=25, alpha=1)

        ax.set_xlabel(f"{self.sample2.sample} {column}")
        ax.set_ylabel(f"{self.sample1.sample} {column}")

        plt.legend(["Uncalled Nucleotides", "Frag-MaP Sites"])

        return fig, ax



"""Performs low SHAPE, low Shannon entropy analysis

Citation:
    Siegfried, N., Busan, S., Rice, G. et al. RNA motif discovery by SHAPE and
        mutational profiling (SHAPE-MaP). Nat Methods 11, 959-965 (2014).
        https://doi.org/10.1038/nmeth.3029

Typical usage example:

import rnavigate as rnav
my_sample = rnav.Sample(
    sample="example sample",
    shapemap="my_shape_profile.txt",
    pairprob="pairing_probabilities.txt",
    ss="MFE_structure.ct"
    )
lowss_sample = rnav.analysis.LowSS(my_sample)
plot = lowss_sample.plot_lowss()
plot.save("lowss_figure.svg")
"""

import numpy as np
from rnavigate import Sample, plots, data


class LowSS(Sample):
    """Creates a new RNAvigate Sample which computes and displays Low SHAPE,
    low Shannon entropy regions (LowSS) given a sample containing SHAPE
    reactivities, pairing probabilities, and MFE structure.

    Methods:
        __init__: performs the analysis
        plot_lowss: displays the result and returns plot object

    Attributes:
        sample (str): the new label for this Sample's data on plots
        parent (rnavigate.Sample): the sample from which data is retrieved
        window (int): size of the windows, must be odd
        median_shape (float): global median SHAPE reactivity
        median_entropy (float): global median Shannon entropy
        data (dictionary): dictionary of data keyword: Data objects, keys are:
            "structure" (rnav.data.SecondaryStructure)
                copy of provided MFE structure
            "shapemap" (rnav.data.SHAPEMaP)
                copy of provided SHAPE-MaP data aligned to "structure"
            "pairprob" (rnav.data.PairingProbability)
                copy of pairing probabilities aligned to "structure"
            "entropies" (rnav.data.Profile)
                Profile of Shannon entropies calculated from "pairprob"
            "lowSS" (rnav.data.Annotations)
                annotations defining low SHAPE, low Shannon entropy regions
    """

    def __init__(
        self,
        sample,
        window=55,
        shapemap="shapemap",
        pairprob="pairprob",
        structure="ss",
    ):
        """Perform Low SHAPE, low Shannon entropy analysis given a sample with
        1) reactivities 2) MFE structure 3) pairing probabilities.

        Required arguments:
            sample (rnav.Sample)
                sample with shapemap, pairing probabilities and MFE structure

        Optional arguments:
            window (integer)
                Window size for calculating median SHAPE and Shannon entropy
                Defaults to 55
            shapemap (data keyword)
                data keyword that points to SHAPE-MaP data
                defaults to "shapemap"
            pairprob (data keyword)
                data keyword that points to pairing probabilities data
                defaults to "pairprob"
            structure (data keyword)
                data keyword that points to MFE structure data
                defaults to "ss"
        """
        # store values
        structure = sample.get_data(structure, data.SecondaryStructure)
        shapemap = sample.get_data(shapemap, data.SHAPEMaP)
        pairprob = sample.get_data(pairprob, data.PairingProbability)
        structure = structure.get_aligned_data(structure.null_alignment)
        shapemap = shapemap.get_aligned_data(
            data.SequenceAlignment(shapemap, structure)
        )
        pairprob = pairprob.get_aligned_data(
            data.SequenceAlignment(pairprob, structure)
        )
        entropy = pairprob.get_entropy_profile()
        super().__init__(
            sample=f"{sample.sample}: lowSS",
            structure=structure,
            shapemap=shapemap,
            pairprob=pairprob,
            entropy=entropy,
        )
        self.parent = sample
        self.window = window
        self.median_shape = shapemap.data["Norm_profile"].median(skipna=True)
        self.median_entropy = entropy.data["Entropy"].median(skipna=True)
        self.reset_window()

    def reset_lowss(self, maximum_shape=None, maximum_entropy=0.08):
        """Generates an annotation of lowSS regions. Stored as self.lowSS

        Args:
            maximum_shape (float, optional): maximum normalized SHAPE
                reactivity to be called lowSS. Defaults to median reactivity.
            maximum_entropy (float, optional): maximum shannon entropy to be
                called lowSS. Defaults to 0.8.
        """
        if maximum_shape is None:
            maximum_shape = self.median_shape
        entropies = self.data["entropy"].data["Windowed_entropy"]
        shapes = self.data["shapemap"].data["Windowed_profile"]
        is_lowss = (entropies < maximum_entropy) & (shapes < maximum_shape)
        self.set_data(
            overwrite=True,
            data_keyword="lowSS",
            inputs=data.Annotation.from_boolean_array(
                values=is_lowss,
                window=self.window,
                annotation_type="spans",
                color="grey",
                sequence=self.data["structure"].sequence,
                name="lowSS",
            ),
        )

    def reset_window(self, window=None):
        """Resets the window size and recalculates windowed SHAPE reactivities
        and shannon entropies and lowSS region annotations.

        Optional arguments:
            window (integer)
                window size, must be an odd integer
                Defaults to value of window attribute
        """
        if window is None:
            window = self.window
        else:
            self.window = window
        # Calculate windowed median SHAPE and entropies
        self.data["shapemap"].calculate_windows(
            column="Norm_profile",
            window=window,
            method="median",
            new_name="Windowed_profile",
            minimum_points=1,
        )
        self.data["entropy"].calculate_windows(
            column="Entropy",
            window=window,
            method="median",
            new_name="Windowed_entropy",
            minimum_points=1,
        )
        self.reset_lowss()

    def plot_lowss(self, region=None, colorbars=True):
        """Visualize LowSS analysis over the given region.

        Optional arguments:
            region (integer or list of 2 integers)
                lowSS region number or start and end positions to plot.
                For region numbers, +/- 150 nts are shown.
                Defaults to entire sequence.
            colorbars (True or False)
                whether to plot colorbars for pairing probability

        Returns:
            rnavigate.plots.AP: LowSS visualization
        """
        shapemap = self.data["shapemap"]
        structure = self.data["structure"]
        pairprob = self.data["pairprob"]
        entropy = self.data["entropy"]
        lowss = self.data["lowSS"]
        # show entire RNA if region is not provided
        if region is None:
            start = 1
            stop = structure.length
            region = [start, stop]
        # if region is an integer, get that lowSS region +/- 150 nts
        elif isinstance(region, int):
            region = lowss[region - 1]
            start, stop = region
            start = max(start - 150, 1)
            stop = min(stop + 150, structure.length)
        # else region should be a list of 2 integers, start and end.
        else:
            start, stop = region
        region_length = stop - start + 1
        # create arc plot instance
        plot = plots.AP(1, region_length, cols=1, rows=1, region=region)
        ax = plot.axes[0, 0]
        x_values = np.arange(start, stop + 1)
        # plot median SHAPE on secondary ax
        prof_ax = ax.twinx()
        prof_ax.set_ylim(-3, 1)
        prof_ax.set_yticks([0.0, 0.4, 0.85])
        prof_ax.fill_between(
            x=x_values,
            y1=[self.median_shape] * region_length,
            y2=shapemap.data["Windowed_profile"][start - 1 : stop],
            fc="0.3",
            lw=0,
        )
        plots.adjust_spines(prof_ax, ["left"])
        plots.clip_spines(prof_ax, ["left"])

        # plot median entropy on second secondary ax
        ent_ax = ax.twinx()
        ent_ax.set_ylim(-1.5, 1.5)
        ent_ax.set_yticks(ticks=[0, 0.08, 0.5])
        ent_ax.fill_between(
            x=x_values,
            y1=[0.08] * region_length,
            y2=entropy.data["Windowed_entropy"][start - 1 : stop],
            fc="C1",
            lw=0,
        )
        plots.adjust_spines(ent_ax, ["left"])
        plots.clip_spines(ent_ax, ["left"])

        # add ss and pairing probabilities track
        plot.plot_data(
            sequence=structure,
            structure=structure,
            interactions=pairprob,
            label=self.sample,
            seqbar=False,
            title=False,
            annotations=[lowss],
            annotation_mode="bar",
        )

        # Place Track Labels
        ax.set_title(
            f"{self.sample}\n{start} - {stop}", loc="left", fontdict={"fontsize": 16}
        )
        ax.text(
            1.002,
            7 / 8,
            "SHAPE\nReactivity",
            fontsize=12,
            transform=ax.transAxes,
            va="center",
        )
        ax.text(
            1.002,
            5 / 8,
            "Shannon\nEntropy",
            fontsize=12,
            transform=ax.transAxes,
            va="center",
        )
        ax.text(
            1.002,
            3 / 8,
            "Secondary\nStructure",
            transform=ax.transAxes,
            fontsize=12,
            va="center",
        )
        ax.text(
            1.002,
            1 / 8,
            "Pairing\nProbability",
            transform=ax.transAxes,
            va="center",
            fontsize=12,
        )

        # Place region labels
        for i, (lssr_start, lssr_stop) in enumerate(lowss):
            middle = (lssr_stop + lssr_start) / 2
            if start < middle < stop:
                ax.text(
                    x=middle, y=550, s=str(i + 1), ha="center", va="center", fontsize=12
                )

        ax.set_ylim([-305, 915])
        ax.set_xticks(ticks=[x for x in range(500, stop + 1, 500) if x > start])
        ax.set_xticks(
            ticks=[x for x in range(100, stop + 1, 100) if x > start], minor=True
        )
        ax.tick_params(axis="x", which="major", labelsize=12)
        plots.adjust_spines(ax, ["bottom"])

        # set figure size so that 200 ax units == 1 inch
        plot.set_figure_size(height_ax_rel=1 / 200, width_ax_rel=1 / 200)
        if colorbars:
            plot.plot_colorbars()
        return plot



"""Plot nucleotide distribution of a profile."""

import numpy as np
import seaborn as sns
from rnavigate import plots, styles


class NucleotideDistribution(plots.Plot):
    """Plot nucleotide distribution of a profile.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    sharex : bool, optional
        Whether to share the x-axis between plots.
    cols : int, optional
        Number of columns in the plot.
    **plot_kwargs
        Keyword arguments passed to the plot function.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples, **plot_kwargs):
        """Initialize the plot."""
        super().__init__(num_samples, sharex=True, cols=1, **plot_kwargs)

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=2,
        height_ax_in=2,
        height_gap_in=0.2,
        width_gap_in=0.4,
        top_in=1,
        bottom_in=1,
        left_in=1,
        right_in=1,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float, optional
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, optional
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, defaults to 2
            Width of the axes in inches.
        height_ax_in : float, defaults to 2
            Height of the axes in inches.
        height_gap_in : float, defaults to 0.2
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.4
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Top margin in inches.
        bottom_in : float, defaults to 1
            Bottom margin in inches.
        left_in : float, defaults to 1
            Left margin in inches.
        right_in : float, defaults to 1
            Right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(self, profile, label, column=None, normalize=None, ax=None):
        """Plot data to the current (or specified) axes.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        label : str
            label for the y-axis.
        column : str, optional
            Column of `profile.data` to plot.
        normalize : dict, optional
            Keyword arguments passed to `profile.normalize`.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot to. If not specified, the current axes is used.
        """
        profile = profile.copy()
        if ax is None:
            ax = self.get_ax()
            self.i += 1
        if column is None:
            column = profile.metric
        if normalize is not None:
            profile.normalize(**normalize)
        data = profile.data
        nt_idx = data[column] > 0
        sns.kdeplot(
            ax=ax,
            data=data.loc[nt_idx],
            x=np.log10(data.loc[nt_idx, column]),
            hue="Sequence",
            hue_order=["A", "U", "C", "G"],
            palette={nt: styles.get_nt_color(nt) for nt in "AUGC"},
            common_norm=False,
        )
        ax.set(
            xlim=(-2.5, 1.5),
            xticks=[-2, -1, 0, 1],
            xticklabels=[0.01, 0.1, 1, 10],
            xlabel="Normalized profile (log scale)",
            yticks=[],
            ylabel=label,
        )
        ax.axvspan(-1, 0, color="lightgrey", alpha=0.4, ec="none")



"""Plot linear regression scatter plots."""

from scipy import stats
import seaborn as sns
import numpy as np
from rnavigate import plots


class LinReg(plots.Plot):
    """Plot a linear regression scatter plot, pairwise, between profiles.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    scale : str, optional
        Scale of the plot, either 'linear' or 'log'. The default is 'linear'.
    regression : 'pearson' or 'spearman', Defaults to 'pearson'
        Type of regression to perform.
    kde : bool, optional
        Whether to plot a kernel density estimate instead of a scatter plot.
        The default is False.
    region : 'all' or tuple of 2 integers, defaults to 'all'
        Start and end positions of the region to plot. If 'all', plot the
        entire profile.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure of the plot.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Axes of the plot.
    length : int
        Number of samples to plot.
    lims : list of 2 floats
        Limits of the plot applied to all x and y axes.
    profiles : list of numpy.ndarray
        Each sample's per-nucleotide values over the region and column of interest.
    colors : list of numpy.ndarray
        A color for each nucleotide in the region applied to the scatter plot.
    labels : list of str
        A label for each sample.
    scale : 'linear' or 'log'
        Scale of the plot axes.
    kde : bool
        Whether to plot a kernel density estimate instead of a scatter plot.
    regression : function
        Regression function to use.
    region : tuple of 2 integers
        Start and end positions of the region to plot.
    """

    def __init__(
        self, num_samples, scale="linear", regression="pearson", kde=False, region="all"
    ):
        """Initialize the linear regression plot."""
        super().__init__(num_samples, rows=num_samples - 1, cols=num_samples - 1)
        self.region = region
        linreg_axes = []
        for row in range(num_samples - 1):
            for col in range(row, num_samples - 1):
                linreg_axes.append(self.axes[row, col])
        self.axes[0, 0].get_shared_x_axes().join(*linreg_axes)
        self.axes[0, 0].get_shared_y_axes().join(*linreg_axes)
        self.lims = [0, 0]
        self.scale = scale
        self.kde = kde
        if regression == "pearson":
            self.regression = stats.pearsonr
        elif regression == "spearman":
            self.regression = stats.spearmanr
        self.profiles = []
        self.colors = []
        self.labels = []

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=2,
        height_ax_in=2,
        height_gap_in=0.3,
        width_gap_in=0.3,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to None
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, defaults to None
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, defaults to 2
            Width of the axes in inches.
        height_ax_in : float, defaults to 2
            Height of the axes in inches.
        height_gap_in : float, defaults to 0.3
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.3
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Top margin of the figure in inches.
        bottom_in : float, defaults to 0.5
            Bottom margin of the figure in inches.
        left_in : float, defaults to 0.5
            Left margin of the figure in inches.
        right_in : float, defaults to 0.5
            Right margin of the figure in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(
        self, structure, profile, annotations, label, column=None, colors="sequence"
    ):
        """Add profile data. If all samples have been added, plot the regression.

        Parameters
        ----------
        structure : Structure
            Structure object.
        profile : Profile
            Profile object.
        annotations : Annotations
            Annotations object.
        label : str
            Label for the sample.
        column : str, optional
            Column of the profile to plot. The default is None.
        colors : str, optional
            Color scheme to use. The default is 'sequence'.
        """
        if self.region == "all":
            start, end = 1, profile.length
        else:
            start, end = self.region
        if column is None:
            column = profile.metric
        self.labels.append(label)
        colors, colormap = profile.get_colors(
            colors,
            profile=profile,
            structure=structure,
            annotations=annotations,
        )
        self.colors.append(colors[start - 1 : end])
        self.add_colorbar_args(colormap)
        self.profiles.append(profile.data[column].to_numpy(copy=True)[start - 1 : end])
        if len(self.profiles) == self.length:
            for row in range(self.length - 1):
                for col in range(self.length - 1):
                    if row <= col:
                        self.plot_regression(i=row, j=col)
            self.finalize()

    def finalize(self):
        """Finalize the plot by formatting the axes and adding labels."""
        # change the plotting buffer for linear scale
        if self.scale == "log" and self.kde:
            # change the tick labels to 'fake' a log scale plot
            ticks = []
            for power in range(-5, 5):
                power = 10**power
                if self.lims[0] > power or power > self.lims[1]:
                    continue
                ticks.append(power)
                # 10^n formatted tick labels
            for i in range(len(self.axes)):
                self.axes[i, i].set(xticks=ticks, yticks=ticks)
        if self.scale == "linear":
            buffer = 0.05 * (self.lims[1] - self.lims[0])
            self.lims[0] -= buffer
            self.lims[1] += buffer
            self.axes[0, 0].set(ylim=self.lims, xlim=self.lims)
        # format axis spines and labels
        ticks = self.axes[0, 0].get_yticks()[1:-1]
        for row in range(self.length - 1):
            for col in range(self.length - 1):
                ax = self.axes[row, col]
                if row <= col:
                    sns.despine(ax=ax)
                    ax.set_xticks(ticks)
                    ax.set_yticks(ticks)
                else:
                    self.axes[row, col].set_axis_off()
                if row < col:
                    ax.set_xticklabels([" "] * len(ticks))
                    ax.set_yticklabels([" "] * len(ticks))
                if row == 0:
                    ax.annotate(
                        self.labels[col + 1],
                        xy=(0.5, 1),
                        xytext=(0, 5),
                        xycoords="axes fraction",
                        textcoords="offset points",
                        size="large",
                        ha="center",
                        va="baseline",
                    )
                if col == (self.length - 2):
                    ax.annotate(
                        self.labels[row],
                        xy=(1, 0.5),
                        xytext=(5, 0),
                        rotation=-90,
                        xycoords="axes fraction",
                        textcoords="offset points",
                        size="large",
                        ha="left",
                        va="center",
                    )

    def plot_regression(self, i, j):
        """Plot a linear regression between two profiles.

        Profiles must already have been added using `plot_data`.

        Parameters
        ----------
        i : int
            Index of the first profile.
        j : int
            Index of the second profile.
        """
        ax = self.axes[i, j]
        p1 = self.profiles[j + 1]
        p2 = self.profiles[i]
        colors = self.colors[i]

        # ax.plot([0, 1], [0, 1], color='black')
        if self.scale == "log":
            notNans = (p1 > 0) & (p2 > 0)
            p1 = p1[notNans]
            p2 = p2[notNans]
            colors = colors[notNans]
            r_value, _ = self.regression(np.log10(p2), np.log10(p1))
        if self.scale == "linear":
            notNans = ~np.isnan(p1) & ~np.isnan(p2)
            p1 = p1[notNans]
            p2 = p2[notNans]
            colors = colors[notNans]
            r_value, _ = self.regression(p2, p1)
        self.lims[0] = min([self.lims[0], min(p1), min(p2)])
        self.lims[1] = max([self.lims[1], max(p1), max(p2)])
        ax.text(
            0.1,
            0.95,
            f"r = {r_value:.2f}",
            transform=ax.transAxes,
            ha="left",
            va="top",
            bbox=dict(fc="white", alpha=0.5, ec="black"),
        )
        if self.kde:
            if self.scale == "log":
                sns.kdeplot(
                    ax=ax,
                    x=p1,
                    y=p2,
                    fill=True,
                    log_scale=True,
                    levels=np.arange(1, 11) / 10,
                )
            elif self.scale == "linear":
                sns.kdeplot(ax=ax, x=p1, y=p2, fill=True, levels=np.arange(1, 11) / 10)
        else:
            ax.scatter(p2, p1, c=colors, marker=".")
            ax.set(xscale=self.scale, yscale=self.scale)
